@@ -8,9 +8,11 @@ package require Pextlib
 
 # Globals
 set full_reindex 0
+set permit_error 0
 set stats(total) 0
 set stats(failed) 0
 set stats(skipped) 0
+set extended_mode 0
 array set ui_options        [list ports_no_old_index_warning 1]
 array set global_options    [list]
 array set global_variations [list]
@@ -22,11 +24,13 @@ mportinit ui_options global_options global_variations
 # Standard procedures
 proc print_usage args {
     global argv0
-    puts "Usage: $argv0 \[-adf\] \[-p plat_ver_arch\] \[-o output directory\] \[directory\]"
-    puts "-o:\tOutput all files to specified directory"
+    puts "Usage: $argv0 \[-dfe\] \[-o output directory\] \[-p plat_ver_\[cxxlib_\]arch\] \[directory\]"
     puts "-d:\tOutput debugging information"
     puts "-f:\tDo a full re-index instead of updating"
+    puts "-e:\tExit code indicates if ports failed to parse"
+    puts "-o:\tOutput all files to specified directory"
     puts "-p:\tPretend to be on another platform"
+    puts "-x:\tInclude extra (optional) information in the PortIndex, like variant description and port notes."
 }
 
 proc _read_index {idx} {
@@ -84,7 +88,7 @@ proc _open_port {portinfo_name portdir absportdir port_options_name {subport {}}
     # Make sure $prefix expands to '${prefix}' so that the PortIndex is
     # portable across prefixes, see https://trac.macports.org/ticket/53169 and
     # https://trac.macports.org/ticket/17182.
-    try -pass_signal {
+    macports_try -pass_signal {
         set macports::prefix {${prefix}}
         if {$subport eq {}} {
             set interp [mportopen file://$absportdir $port_options]
@@ -114,30 +118,34 @@ proc pindex {portdir} {
     set portfile [file join $absportdir Portfile]
     # try to reuse the existing entry if it's still valid
     if {$full_reindex != 1 && [info exists qindex($qname)]} {
-        try -pass_signal {
+        macports_try -pass_signal {
             set mtime [file mtime $portfile]
             if {$oldmtime >= $mtime} {
                 lassign [_read_index $qname] name len line
-                _write_index $name $len $line
-                incr stats(skipped)
-
-                if {[info exists ui_options(ports_debug)]} {
-                    puts "Reusing existing entry for $portdir"
-                }
-
-                # also reuse the entries for its subports
                 array set portinfo $line
-                if {![info exists portinfo(subports)]} {
+
+                # reuse entry if it was made from the same portdir
+                if {[info exists portinfo(portdir)] && $portinfo(portdir) eq $portdir} {
+                    _write_index $name $len $line
+                    incr stats(skipped)
+
+                    if {[info exists ui_options(ports_debug)]} {
+                        puts "Reusing existing entry for $portdir"
+                    }
+
+                    # also reuse the entries for its subports
+                    if {![info exists portinfo(subports)]} {
+                        return
+                    }
+                    foreach sub $portinfo(subports) {
+                        _write_index {*}[_read_index [string tolower $sub]]
+                        incr stats(skipped)
+                    }
+
                     return
                 }
-                foreach sub $portinfo(subports) {
-                    _write_index {*}[_read_index [string tolower $sub]]
-                    incr stats(skipped)
-                }
-
-                return
             }
-        } catch {{*} eCode eMessage} {
+        } on error {} {
             ui_warn "Failed to open old entry for ${portdir}, making a new one"
             if {[info exists ui_options(ports_debug)]} {
                 puts "$::errorInfo"
@@ -146,7 +154,7 @@ proc pindex {portdir} {
     }
 
     incr stats(total)
-    try -pass_signal {
+    macports_try -pass_signal {
         _open_port portinfo $portdir $absportdir port_options
         puts "Adding port $portdir"
 
@@ -162,17 +170,17 @@ proc pindex {portdir} {
         }
         foreach sub $portinfo(subports) {
             incr stats(total)
-            try -pass_signal {
+            macports_try -pass_signal {
                 _open_port portinfo $portdir $absportdir port_options $sub
                 puts "Adding subport $sub"
 
                 _write_index_from_portinfo portinfo yes
-            } catch {{*} eCode eMessage} {
+            } on error {eMessage} {
                 puts stderr "Failed to parse file $portdir/Portfile with subport '${sub}': $eMessage"
                 incr stats(failed)
             }
         }
-    } catch {{*} eCode eMessage} {
+    } on error {eMessage} {
         puts stderr "Failed to parse file $portdir/Portfile: $eMessage"
         incr stats(failed)
     }
@@ -197,14 +205,43 @@ for {set i 0} {$i < $argc} {incr i} {
                 set platlist [split [lindex $argv $i] _]
                 set os_platform [lindex $platlist 0]
                 set os_major [lindex $platlist 1]
-                set os_arch [lindex $platlist 2]
+                if {[llength $platlist] > 3} {
+                    set cxx_stdlib [lindex $platlist 2]
+                    switch -- $cxx_stdlib {
+                        libcxx {
+                            set cxx_stdlib libc++
+                        }
+                        libstdcxx {
+                            set cxx_stdlib libstdc++
+                        }
+                        default {
+                            puts stderr "Unknown C++ standard library: $cxx_stdlib (use libcxx or libstdcxx)"
+                            print_usage
+                            exit 1
+                        }
+                    }
+                    set os_arch [lindex $platlist 3]
+                } else {
+                    if {$os_platform eq "macosx"} {
+                        if {$os_major < 10} {
+                            set cxx_stdlib libstdc++
+                        } else {
+                            set cxx_stdlib libc++
+                        }
+                    }
+                    set os_arch [lindex $platlist 2]
+                }
                 if {$os_platform eq "macosx"} {
-                    lappend port_options os.subplatform $os_platform os.universal_supported yes
+                    lappend port_options os.subplatform $os_platform os.universal_supported yes cxx_stdlib $cxx_stdlib
                     set os_platform darwin
                 }
                 lappend port_options os.platform $os_platform os.major $os_major os.arch $os_arch
             } elseif {$arg eq "-f"} { # Completely rebuild index
                 set full_reindex 1
+            } elseif {$arg eq "-x"} { # Build extended portindex (include extra information , eg.: notes, variant description, conflicts etc.)
+                set extended_mode 1
+            } elseif {$arg eq "-e"} { # Non-zero exit code on errors
+                set permit_error 1
             } else {
                 puts stderr "Unknown option: $arg"
                 print_usage
@@ -266,20 +303,30 @@ if {[file isfile $outpath] && [file isfile ${outpath}.quick]} {
 set tempportindex [mktemp "/tmp/mports.portindex.XXXXXXXX"]
 set fd [open $tempportindex w]
 set save_prefix ${macports::prefix}
-foreach key {categories depends_fetch depends_extract depends_build \
-             depends_lib depends_run depends_test description epoch homepage \
-             long_description maintainers name platforms revision variants \
-             version portdir replaced_by license installs_libs conflicts} {
+
+# keys for a normal portindex
+foreach key {categories depends_fetch depends_extract depends_patch \
+             depends_build depends_lib depends_run depends_test \
+             description epoch homepage long_description maintainers \
+             name platforms revision variants version portdir \
+             replaced_by license installs_libs conflicts known_fail} {
     set keepkeys($key) 1
 }
 
+# additional keys for extended portindex (with extra information)
+if {$extended_mode eq 1 } {
+    foreach key {vinfo notes} {
+        set keepkeys($key) 1
+    }
+}
+
 set exit_fail 0
 try {
     mporttraverse pindex $directory
-} catch {{POSIX SIG SIGINT} eCode eMessage} {
+} trap {POSIX SIG SIGINT} {} {
     puts stderr "SIGINT received, terminating."
     set exit_fail 1
-} catch {{POSIX SIG SIGTERM} eCode eMessage} {
+} trap {POSIX SIG SIGTERM} {} {
     puts stderr "SIGTERM received, terminating."
     set exit_fail 1
 } finally {
@@ -299,3 +346,7 @@ puts "\nTotal number of ports parsed:\t$stats(total)\
       \nPorts successfully parsed:\t[expr {$stats(total) - $stats(failed)}]\
       \nPorts failed:\t\t\t$stats(failed)\
       \nUp-to-date ports skipped:\t$stats(skipped)\n"
+
+if {${permit_error} && $stats(failed) > 0} {
+    exit 2
+}
