@@ -1,8 +1,8 @@
 /*
  * entry.c
- * $Id: entry.c 28029 2007-08-18 15:59:59Z sfiera@macports.org $
  *
  * Copyright (c) 2007 Chris Pickel <sfiera@macports.org>
+ * Copyright (c) 2012 The MacPorts Project
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -35,6 +35,8 @@
 #include <tcl.h>
 #include <sqlite3.h>
 
+#include <cregistry/util.h>
+
 #include "entry.h"
 #include "entryobj.h"
 #include "registry.h"
@@ -66,6 +68,7 @@ void delete_entry(ClientData clientData) {
     entry->proc = NULL;
 }
 
+/*
 static int obj_to_entry(Tcl_Interp* interp, reg_entry** entry, Tcl_Obj* obj,
         reg_error* errPtr) {
     reg_entry* result = get_entry(interp, Tcl_GetString(obj), errPtr);
@@ -77,7 +80,6 @@ static int obj_to_entry(Tcl_Interp* interp, reg_entry** entry, Tcl_Obj* obj,
     }
 }
 
-/*
 static int list_obj_to_entry(Tcl_Interp* interp, reg_entry*** entries,
         const Tcl_Obj** objv, int objc, reg_error* errPtr) {
     return recast(interp, (cast_function*)obj_to_entry, NULL, (void***)entries,
@@ -112,7 +114,7 @@ static int entry_create(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[]) {
                 variants, epoch, &error);
         if (entry != NULL) {
             Tcl_Obj* result;
-            if (entry_to_obj(interp, &result, entry, &error)) {
+            if (entry_to_obj(interp, &result, entry, NULL, &error)) {
                 Tcl_SetObjResult(interp, result);
                 return TCL_OK;
             }
@@ -132,7 +134,8 @@ static int entry_delete(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[]) {
     if (objc != 3) {
         Tcl_WrongNumArgs(interp, 1, objv, "delete entry");
         return TCL_ERROR;
-    } if (reg == NULL) {
+    }
+    if (reg == NULL) {
         return TCL_ERROR;
     } else {
         reg_error error;
@@ -144,19 +147,37 @@ static int entry_delete(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[]) {
         if (!reg_entry_delete(entry, &error)) {
             return registry_failed(interp, &error);
         }
+        /* We want to delete the command but not free the entry object,
+           since we may want to restore it (delete_entry_list will free
+           it if the transaction is not rolled back). So remove the
+           command's associated deleteProc first. The reg_entry_free
+           below will take care of it if there is no transaction in
+           progress. */
+        Tcl_CmdInfo info;
+        if (Tcl_GetCommandInfo(interp, Tcl_GetString(objv[2]), &info)) {
+            info.deleteProc = NULL;
+            Tcl_SetCommandInfo(interp, Tcl_GetString(objv[2]), &info);
+        }
         /* if there's a transaction going on, record this entry in a list so we
          * can roll it back if necessary
          */
         list_handle = Tcl_GetAssocData(interp, "registry::deleted", NULL);
         if (list_handle) {
             entry_list* list = *list_handle;
-            *list_handle = malloc(sizeof(entry_list*));
-            (*list_handle)->entry = entry;
-            (*list_handle)->next = list;
+            *list_handle = malloc(sizeof(entry_list));
+            if (*list_handle) {
+                (*list_handle)->entry = entry;
+                (*list_handle)->next = list;
+            } else {
+                reg_entry_free(entry);
+                return TCL_ERROR;
+            }
         } else {
             reg_entry_free(entry);
         }
         Tcl_DeleteCommand(interp, Tcl_GetString(objv[2]));
+        /* set flag so that the db will be vacuumed when we close it */
+        Tcl_SetAssocData(interp, "registry::needs_vacuum", NULL, (ClientData)1);
         return TCL_OK;
     }
 }
@@ -185,7 +206,7 @@ static int entry_open(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[]) {
                 variants, epoch, &error);
         if (entry != NULL) {
             Tcl_Obj* result;
-            if (entry_to_obj(interp, &result, entry, &error)) {
+            if (entry_to_obj(interp, &result, entry, NULL, &error)) {
                 Tcl_SetObjResult(interp, result);
                 return TCL_OK;
             }
@@ -202,7 +223,7 @@ static int entry_open(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[]) {
  */
 static int entry_close(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[]) {
     if (objc != 3) {
-        Tcl_WrongNumArgs(interp, 1, objv, "delete entry");
+        Tcl_WrongNumArgs(interp, 1, objv, "close entry");
         return TCL_ERROR;
     } else {
         reg_error error;
@@ -226,6 +247,7 @@ static strategy_type strategies[] = {
     { "-exact",  reg_strategy_exact },
     { "-glob",   reg_strategy_glob },
     { "-regexp", reg_strategy_regexp },
+    { "-null",   reg_strategy_null },
     { "--",      reg_strategy_exact },
     { NULL, 0 }
 };
@@ -235,76 +257,95 @@ static strategy_type strategies[] = {
  *
  * Searches the registry for ports for which each key's value is equal to the
  * given value. To find all ports, call `entry search` with no key-value pairs.
- * Can be given an option of -exact, -glob, or -regexp to specify the matching
- * strategy; defaults to exact.
+ * Can be given an option of -exact, -glob, -regexp or -null to specify the
+ * matching strategy; defaults to exact.
  */
-static int entry_search(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[]) {
-    int i;
-    reg_registry* reg = registry_for(interp, reg_attached);
-    if ((objc > 2) && ((Tcl_GetString(objv[2])[0] == '-')
-                ? (objc % 2 == 0) : (objc % 2 == 1))) {
-        Tcl_WrongNumArgs(interp, 2, objv, "search ?options? ?key value ...?");
+static int entry_search(Tcl_Interp* interp, int objc, Tcl_Obj* CONST *objv) {
+    reg_registry *reg = registry_for(interp, reg_attached);
+    if (!reg) {
         return TCL_ERROR;
-    } else if (reg == NULL) {
-        return TCL_ERROR;
-    } else {
-        char** keys;
-        char** vals;
-        int key_count = objc/2 - 1;
-        reg_entry** entries;
-        reg_error error;
-        int entry_count;
-        int start;
-        int strategy;
-        /* try to use strategy */
-        if (objc > 2 && Tcl_GetString(objv[2])[0] == '-') {
-            int strat_index;
-            if (Tcl_GetIndexFromObjStruct(interp, objv[2], strategies,
-                        sizeof(strategy_type), "option", 0, &strat_index)
-                    == TCL_ERROR) {
-                return TCL_ERROR;
+    }
+    const char **keys = malloc(objc * sizeof(char *));
+    const char **vals = malloc(objc * sizeof(char *));
+    int *strats = malloc(objc * sizeof(int));
+    if (!keys || !vals || !strats) {
+        goto cleanup_error;
+    }
+    int key_count = 0;
+
+    /* Skip the "registry::entry" "search" bit */
+    objv += 2;
+    objc -= 2;
+
+    for (int i = 0; i < objc; ++key_count /* i in incremented in the loop */) {
+        int key_index, strategy_index;
+
+        /* Grab the key */
+        if (Tcl_GetIndexFromObj(interp, objv[i], entry_props, "search key",
+            0, &key_index) != TCL_OK) {
+            goto cleanup_error;
+        }
+        keys[key_count] = entry_props[key_index];
+        if (objc <= ++i) {
+            goto wrong_num_args; /* We're missing a strategy or value */
+        }
+
+        /* Set the default strategy, in case we don't find one */
+        strats[key_count] = reg_strategy_exact;
+        /* Try to grab the strategy */
+        if (Tcl_GetIndexFromObjStruct(interp, objv[i], strategies,
+                sizeof(strategy_type), "option", 0, &strategy_index) !=
+                TCL_ERROR) {
+            strats[key_count] = strategies[strategy_index].strategy;
+            if (objc <= ++i && strats[key_count] != reg_strategy_null) {
+                goto wrong_num_args;
+            } else if (strats[key_count] == reg_strategy_null) {
+                /* This doesn't take a value, so start from the top */
+                continue;
             }
-            strategy = strategies[strat_index].strategy;
-            start = 3;
+        }
+
+        /* Grab the value */
+        if (!(vals[key_count] = Tcl_GetString(objv[i++]))) {
+            goto wrong_num_args; /* We're missing a value */
+        }
+    }
+
+    reg_entry **entries;
+    reg_error error;
+    int entry_count;
+    entry_count = reg_entry_search(reg, keys, vals, key_count, strats, &entries,
+        &error);
+    free(keys);
+    free(vals);
+    free(strats);
+
+    if (entry_count >= 0) {
+        int retval;
+        Tcl_Obj* resultObj;
+        Tcl_Obj** objs;
+        if (list_entry_to_obj(interp, &objs, entries, entry_count, &error)){
+            resultObj = Tcl_NewListObj(entry_count, objs);
+            Tcl_SetObjResult(interp, resultObj);
+            free(objs);
+            retval = TCL_OK;
         } else {
-            strategy = reg_strategy_exact;
-            start = 2;
-        }
-        /* ensure that valid search keys were used */
-        for (i=start; i<objc; i+=2) {
-            int index;
-            if (Tcl_GetIndexFromObj(interp, objv[i], entry_props, "search key",
-                        0, &index) != TCL_OK) {
-                return TCL_ERROR;
-            }
-        }
-        keys = malloc(key_count * sizeof(char*));
-        vals = malloc(key_count * sizeof(char*));
-        for (i=0; i<key_count; i+=1) {
-            keys[i] = Tcl_GetString(objv[2*i+start]);
-            vals[i] = Tcl_GetString(objv[2*i+start+1]);
-        }
-        entry_count = reg_entry_search(reg, keys, vals, key_count,
-                strategy, &entries, &error);
-        free(keys);
-        free(vals);
-        if (entry_count >= 0) {
-            int retval;
-            Tcl_Obj* resultObj;
-            Tcl_Obj** objs;
-            if (list_entry_to_obj(interp, &objs, entries, entry_count, &error)){
-                resultObj = Tcl_NewListObj(entry_count, objs);
-                Tcl_SetObjResult(interp, resultObj);
-                free(objs);
-                retval = TCL_OK;
-            } else {
-                retval = registry_failed(interp, &error);
-            }
-            free(entries);
-            return retval;
+            retval = registry_failed(interp, &error);
         }
+        free(entries);
+        return retval;
+    } else {
         return registry_failed(interp, &error);
     }
+
+wrong_num_args:
+    Tcl_WrongNumArgs(interp, 2, objv,
+            "search ?key ?options? value ...?");
+cleanup_error:
+    free(keys);
+    free(vals);
+    free(strats);
+    return TCL_ERROR;
 }
 
 /*
@@ -366,14 +407,15 @@ static int entry_imaged(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[]) {
         entry_count = reg_entry_imaged(reg, name, version, revision, variants,
                 &entries, &error);
         if (entry_count >= 0) {
-            Tcl_Obj* resultObj;
             Tcl_Obj** objs;
-            list_entry_to_obj(interp, &objs, entries, entry_count, &error);
-            resultObj = Tcl_NewListObj(entry_count, objs);
-            Tcl_SetObjResult(interp, resultObj);
+            if (list_entry_to_obj(interp, &objs, entries, entry_count, &error)) {
+                Tcl_Obj* resultObj = Tcl_NewListObj(entry_count, objs);
+                Tcl_SetObjResult(interp, resultObj);
+                free(entries);
+                free(objs);
+                return TCL_OK;
+            }
             free(entries);
-            free(objs);
-            return TCL_OK;
         }
         return registry_failed(interp, &error);
     }
@@ -408,14 +450,15 @@ static int entry_installed(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[]){
         }
         entry_count = reg_entry_installed(reg, name, &entries, &error);
         if (entry_count >= 0) {
-            Tcl_Obj* resultObj;
             Tcl_Obj** objs;
-            list_entry_to_obj(interp, &objs, entries, entry_count, &error);
-            resultObj = Tcl_NewListObj(entry_count, objs);
-            Tcl_SetObjResult(interp, resultObj);
+            if (list_entry_to_obj(interp, &objs, entries, entry_count, &error)) {
+                Tcl_Obj* resultObj = Tcl_NewListObj(entry_count, objs);
+                Tcl_SetObjResult(interp, resultObj);
+                free(entries);
+                free(objs);
+                return TCL_OK;
+            }
             free(entries);
-            free(objs);
-            return TCL_OK;
         }
         return registry_failed(interp, &error);
     }
@@ -423,24 +466,37 @@ static int entry_installed(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[]){
 
 
 /*
+ * registry::entry owner filename [cs = true]
+ *
+ * Returns the port that owns the given filename (empty string if none).
  */
 static int entry_owner(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[]) {
     reg_registry* reg = registry_for(interp, reg_attached);
-    if (objc != 3) {
-        Tcl_WrongNumArgs(interp, 2, objv, "path");
+    int cs = 1;
+
+    if ((objc < 3) || (objc > 4)) {
+        Tcl_WrongNumArgs(interp, 2, objv, "path ?cs?");
         return TCL_ERROR;
-    } else if (reg == NULL) {
+    }
+
+    if (objc == 4) {
+        if (Tcl_GetBooleanFromObj(interp, objv[3], &cs) != TCL_OK) {
+            return TCL_ERROR;
+        }
+    }
+
+    if (reg == NULL) {
         return TCL_ERROR;
     } else {
         char* path = Tcl_GetString(objv[2]);
         reg_entry* entry;
         reg_error error;
-        if (reg_entry_owner(reg, path, &entry, &error)) {
+        if (reg_entry_owner(reg, path, cs, &entry, &error)) {
             if (entry == NULL) {
                 return TCL_OK;
             } else {
                 Tcl_Obj* result;
-                if (entry_to_obj(interp, &result, entry, &error)) {
+                if (entry_to_obj(interp, &result, entry, NULL, &error)) {
                     Tcl_SetObjResult(interp, result);
                     return TCL_OK;
                 }
