@@ -1,7 +1,7 @@
 # -*- coding: utf-8; mode: tcl; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- vim:fenc=utf-8:ft=tcl:et:sw=4:ts=4:sts=4
 # portimage.tcl
 #
-# Copyright (c) 2004-2005, 2007-2011, 2014 The MacPorts Project
+# Copyright (c) 2004-2005, 2007-2018 The MacPorts Project
 # Copyright (c) 2004 Will Barton <wbb4@opendarwin.org>
 # Copyright (c) 2002 Apple Inc.
 # All rights reserved.
@@ -88,7 +88,7 @@ proc activate {name {version ""} {revision ""} {variants 0} {optionslist ""}} {
     if {[info exists options(ports_activate_no-exec)]} {
         set noexec $options(ports_activate_no-exec)
     }
-    set rename_list {}
+    set rename_list [list]
     if {[info exists options(portactivate_rename_files)]} {
         set rename_list $options(portactivate_rename_files)
     }
@@ -109,34 +109,50 @@ proc activate {name {version ""} {revision ""} {variants 0} {optionslist ""}} {
         set specifier "${version}_${revision}${variants}"
         set location [$requested location]
 
+        if {[$requested state] eq "installed"} {
+            ui_info "${name} @${specifier} is already active."
+            #registry::entry close $requested
+            return
+        }
+
+        # this shouldn't be possible
+        if { [$requested installtype] ne "image" } {
+            #registry::entry close $requested
+            return -code error "Image error: ${name} @${specifier} not installed as an image."
+        }
+        if {![::file isfile $location]} {
+            #registry::entry close $requested
+            return -code error "Image error: Can't find image file $location"
+        }
+
         # if another version of this port is active, deactivate it first
         set current [registry::entry installed $name]
         foreach i $current {
             if { $specifier ne "[$i version]_[$i revision][$i variants]" } {
                 lappend todeactivate $i
+            } else {
+                #registry::entry close $i
+            }
+        }
+    }
+
+    try {
+        foreach a $todeactivate {
+            if {$noexec || ![registry::run_target $a deactivate [list ports_nodepcheck 1]]} {
+                deactivate $name [$a version] [$a revision] [$a variants] [list ports_nodepcheck 1]
             }
         }
 
-        # this shouldn't be possible
-        if { [$requested installtype] ne "image" } {
-            return -code error "Image error: ${name} @${specifier} not installed as an image."
-        }
-        if {![::file isfile $location]} {
-            return -code error "Image error: Can't find image file $location"
-        }
-        if {[$requested state] eq "installed"} {
-            return -code error "Image error: ${name} @${specifier} is already active."
+        ui_msg "$UI_PREFIX [format [msgcat::mc "Activating %s @%s"] $name $specifier]"
+
+        _activate_contents $requested $rename_list
+    } finally {
+        #registry::entry close $requested
+        foreach a $todeactivate {
+            # may have been closed by deactivate
+            #catch {registry::entry close $a}
         }
     }
-    foreach a $todeactivate {
-        if {$noexec || ![registry::run_target $a deactivate [list ports_nodepcheck 1]]} {
-            deactivate $name [$a version] [$a revision] [$a variants] [list ports_nodepcheck 1]
-        }
-    }
-
-    ui_msg "$UI_PREFIX [format [msgcat::mc "Activating %s @%s"] $name $specifier]"
-
-    _activate_contents $requested $rename_list
 }
 
 # takes a composite version spec rather than separate version,revision,variants
@@ -173,7 +189,23 @@ proc deactivate {name {version ""} {revision ""} {variants 0} {optionslist ""}}
     if { [llength $ilist] == 1 } {
         set requested [lindex $ilist 0]
     } else {
-        throw registry::image-error "Image error: port ${name} is not active."
+        set ilist [_check_registry $name $version $revision $variants 1]
+        if {[llength $ilist] > 0} {
+            ui_info "port ${name} is already inactive"
+            #foreach i $ilist {
+            #    registry::entry close $i
+            #}
+            return
+        } else {
+            set v ""
+            if {$version ne ""} {
+                set v " @${version}"
+                if {$revision ne ""} {
+                    append v _${revision}${variants}
+                }
+            }
+            throw registry::image-error "Image error: port ${name}${v} is not active."
+        }
     }
     # set name again since the one we were passed may not have had the correct case
     set name [$requested name]
@@ -185,30 +217,57 @@ proc deactivate {name {version ""} {revision ""} {variants 0} {optionslist ""}}
         if {$revision ne ""} {
             append v _${revision}${variants}
         }
+        set ilist [_check_registry $name $version $revision $variants 1]
+        foreach inact $ilist {
+            if {$revision ne ""} {
+                set thisv [$inact version]_[$inact revision][$inact variants]
+            } else {
+                set thisv [$inact version]
+            }
+            if {$v eq $thisv} {
+                ui_info "port ${name} @${thisv} is already inactive"
+                #registry::entry close $requested
+                #foreach inact $ilist {
+                #    registry::entry close $inact
+                #}
+                return
+            }
+        }
+        #registry::entry close $requested
+        #foreach inact $ilist {
+        #    registry::entry close $inact
+        #}
         return -code error "Active version of $name is not $v but ${specifier}."
     }
 
     if { [$requested installtype] ne "image" } {
+        #registry::entry close $requested
         return -code error "Image error: ${name} @${specifier} not installed as an image."
     }
     # this shouldn't be possible
     if { [$requested state] ne "installed" } {
+        #registry::entry close $requested
         return -code error "Image error: ${name} @${specifier} is not active."
     }
-	
+
     if {![info exists options(ports_nodepcheck)] || ![string is true -strict $options(ports_nodepcheck)]} {
         set retvalue [registry::check_dependents $requested $force "deactivate"]
         if {$retvalue eq "quit"} {
+            #registry::entry close $requested
             return
         }
     }
 
     ui_msg "$UI_PREFIX [format [msgcat::mc "Deactivating %s @%s"] $name $specifier]"
-	
-    _deactivate_contents $requested [$requested files] $force
+
+    try {
+        _deactivate_contents $requested [$requested files] $force
+    } finally {
+        #registry::entry close $requested
+    }
 }
 
-proc _check_registry {name version revision variants} {
+proc _check_registry {name version revision variants {return_all 0}} {
     global UI_PREFIX
 
     set searchkeys $name
@@ -224,9 +283,12 @@ proc _check_registry {name version revision variants} {
         }
     }
     set ilist [registry::entry imaged {*}$searchkeys]
+    if {$return_all} {
+        return $ilist
+    }
 
     if { [llength $ilist] > 1 } {
-        set portilist {}
+        set portilist [list]
         set msg "The following versions of $name are currently installed:"
         if {[macports::ui_isset ports_noninteractive]} {
             ui_msg "$UI_PREFIX [msgcat::mc $msg]"
@@ -244,9 +306,18 @@ proc _check_registry {name version revision variants} {
             }
         }
         if {[info exists macports::ui_options(questions_singlechoice)]} {
-            set retvalue [$macports::ui_options(questions_singlechoice) $msg "Choice_Q1" $portilist]
-            return [lindex $ilist $retvalue]
+            set retindex [$macports::ui_options(questions_singlechoice) $msg "Choice_Q1" $portilist]
+            set retvalue [lindex $ilist $retindex]
+            #foreach i $ilist {
+            #    if {$i ne $retvalue} {
+            #        registry::entry close $i
+            #    }
+            #}
+            return $retvalue
         }
+        #foreach i $ilist {
+        #    registry::entry close $i
+        #}
         throw registry::invalid "Registry error: Please specify the full version as recorded in the port registry."
     } elseif { [llength $ilist] == 1 } {
         return [lindex $ilist 0]
@@ -311,7 +382,7 @@ proc extract_archive_to_tmpdir {location} {
         if {[catch {cd $extractdir} err]} {
             throw MACPORTS $err
         }
-    
+
         # clagged straight from unarchive... this really needs to be factored
         # out, but it's a little tricky as the places where it's used run in
         # different interpreter contexts with access to different packages.
@@ -336,13 +407,13 @@ proc extract_archive_to_tmpdir {location} {
                         set gzip "gzip"
                         if {[catch {set gzip [macports::findBinary $gzip ${macports::autoconf::gzip_path}]} errmsg] == 0} {
                             ui_debug "Using $gzip"
-                            set unarchive.pipe_cmd "$gzip -d -c ${location} |"
+                            set unarchive.pipe_cmd "$gzip -d -c [macports::shellescape ${location}] |"
                         } else {
                             ui_debug $errmsg
                             throw MACPORTS "No '$gzip' was found on this system!"
                         }
                     } else {
-                        set unarchive.args "-f ${location} ."
+                        set unarchive.args "-f [macports::shellescape ${location}] ."
                     }
                 } else {
                     ui_debug $errmsg
@@ -353,7 +424,9 @@ proc extract_archive_to_tmpdir {location} {
                 global macports::hfscompression
                 # Opportunistic HFS compression. bsdtar will automatically
                 # disable this if filesystem does not support compression.
-                # Have to be root due to <https://trac.macports.org/ticket/56563>.
+                # Don't use if not running as root, due to bugs:
+                # The system bsdtar on 10.15 suffers from https://github.com/libarchive/libarchive/issues/497
+                # Later versions fixed that problem but another remains: https://github.com/libarchive/libarchive/issues/1415 
                 if {${macports::hfscompression} && [getuid] == 0 &&
                         ![catch {macports::binaryInPath bsdtar}] &&
                         ![catch {exec bsdtar -x --hfsCompression < /dev/null >& /dev/null}]} {
@@ -395,13 +468,13 @@ proc extract_archive_to_tmpdir {location} {
                     }
                     if {[catch {set gzip [macports::findBinary $gzip $hint]} errmsg] == 0} {
                         ui_debug "Using $gzip"
-                        set unarchive.pipe_cmd "$gzip -d -c ${location} |"
+                        set unarchive.pipe_cmd "$gzip -d -c [macports::shellescape ${location}] |"
                     } else {
                         ui_debug $errmsg
                         throw MACPORTS "No '$gzip' was found on this system!"
                     }
                 } else {
-                    set unarchive.args "${location}"
+                    set unarchive.args [macports::shellescape ${location}]
                 }
             }
             xar {
@@ -410,7 +483,7 @@ proc extract_archive_to_tmpdir {location} {
                     ui_debug "Using $xar"
                     set unarchive.cmd "$xar"
                     set unarchive.pre_args {-xvpf}
-                    set unarchive.args "${location}"
+                    set unarchive.args [macports::shellescape ${location}]
                 } else {
                     ui_debug $errmsg
                     throw MACPORTS "No '$xar' was found on this system!"
@@ -426,7 +499,7 @@ proc extract_archive_to_tmpdir {location} {
                     } else {
                         set unarchive.pre_args {-o}
                     }
-                    set unarchive.args "${location} -d ."
+                    set unarchive.args "[macports::shellescape ${location}] -d ."
                 } else {
                     ui_debug $errmsg
                     throw MACPORTS "No '$unzip' was found on this system!"
@@ -436,7 +509,7 @@ proc extract_archive_to_tmpdir {location} {
                 throw MACPORTS "Unsupported port archive type '${unarchive.type}'!"
             }
         }
-        
+
         # and finally, reinvent command_exec
         if {${unarchive.pipe_cmd} eq ""} {
             set cmdstring "${unarchive.cmd} ${unarchive.pre_args} ${unarchive.args}"
@@ -444,9 +517,9 @@ proc extract_archive_to_tmpdir {location} {
             set cmdstring "${unarchive.pipe_cmd} ( ${unarchive.cmd} ${unarchive.pre_args} ${unarchive.args} )"
         }
         system $cmdstring
-    } catch {*} {
+    } on error {_ eOptions} {
         ::file delete -force $extractdir
-        throw
+        throw [dict get $eOptions -errorcode] [dict get $eOptions -errorinfo]
     } finally {
         cd $startpwd
     }
@@ -464,6 +537,7 @@ proc _activate_contents {port {rename_list {}}} {
     set location [$port location]
     set imagefiles [$port imagefiles]
     set extracted_dir [extract_archive_to_tmpdir $location]
+    set replaced_by_re "(?i)^[$port name]\$"
 
     set backups [list]
     # This is big and hairy and probably could be done better.
@@ -493,7 +567,7 @@ proc _activate_contents {port {rename_list {}}} {
                     set result [mportlookup [$owner name]]
                     array unset portinfo
                     array set portinfo [lindex $result 1]
-                    if {[info exists portinfo(replaced_by)] && [lsearch -regexp $portinfo(replaced_by) "(?i)^[$port name]\$"] != -1} {
+                    if {[info exists portinfo(replaced_by)] && [lsearch -regexp $portinfo(replaced_by) $replaced_by_re] != -1} {
                         # we'll deactivate the owner later, but before activating our files
                         set todeactivate($owner) yes
                         set owner "replaced"
@@ -520,11 +594,17 @@ proc _activate_contents {port {rename_list {}}} {
                         # we find any files that already exist, or have entries in
                         # the registry
                         if { $owner ne {} && $owner ne $port } {
-                            throw registry::image-error "Image error: $file is being used by the active [$owner name] port.  Please deactivate this port first, or use 'port -f activate [$port name]' to force the activation."
+                            set msg "Image error: $file is being used by the active [$owner name] port.  Please deactivate this port first, or use 'port -f activate [$port name]' to force the activation."
+                            #registry::entry close $owner
+                            throw registry::image-error $msg
                         } elseif { $owner eq {} && ![catch {::file type $file}] } {
-                            throw registry::image-error "Image error: $file already exists and does not belong to a registered port.  Unable to activate port [$port name]. Use 'port -f activate [$port name]' to force the activation."
+                            set msg "Image error: $file already exists and does not belong to a registered port.  Unable to activate port [$port name]. Use 'port -f activate [$port name]' to force the activation."
+                            throw registry::image-error $msg
                         }
                     }
+                    #if {$owner ne {}} {
+                    #    registry::entry close $owner
+                    #}
                 }
 
                 # Split out the filename's subpaths and add them to the
@@ -560,7 +640,7 @@ proc _activate_contents {port {rename_list {}}} {
         # debug output of activate make more sense.
         set files [lsort -increasing -unique $files]
         # handle files that are to be renamed
-        set confirmed_rename_list {}
+        set confirmed_rename_list [list]
         foreach {src dest} $rename_list {
             set index [lsearch -exact -sorted $files $src]
             if {$index != -1} {
@@ -568,7 +648,7 @@ proc _activate_contents {port {rename_list {}}} {
                 lappend confirmed_rename_list $src $dest
             }
         }
-        set rollback_filelist {}
+        set rollback_filelist [list]
 
         registry::write {
             # Activate it, and catch errors so we can roll-back
@@ -591,27 +671,27 @@ proc _activate_contents {port {rename_list {}}} {
                 # here so that this information cannot be inconsistent with the
                 # state of the files on disk.
                 $port state installed
-            } catch {{POSIX SIG SIGINT} eCode eMessage} {
+            } trap {POSIX SIG SIGINT} {_ eOptions} {
                 # Pressing ^C will (often?) print "^C" to the terminal; send
                 # a linebreak so our message appears after that.
                 ui_msg ""
                 ui_msg "Control-C pressed, rolling back, please wait."
                 # can't do it here since we're already inside a transaction
                 set deactivate_this yes
-                throw
-            } catch {{POSIX SIG SIGTERM} eCode eMessage} {
+                throw [dict get $eOptions -errorcode] [dict get $eOptions -errorinfo]
+            } trap {POSIX SIG SIGTERM} {_ eOptions} {
                 ui_msg "SIGTERM received, rolling back, please wait."
                 # can't do it here since we're already inside a transaction
                 set deactivate_this yes
-                throw
-            } catch {*} {
+                throw [dict get $eOptions -errorcode] [dict get $eOptions -errorinfo]
+            } on error {_ eOptions} {
                 ui_debug "Activation failed, rolling back."
                 # can't do it here since we're already inside a transaction
                 set deactivate_this yes
-                throw
+                throw [dict get $eOptions -errorcode] [dict get $eOptions -errorinfo]
             }
         }
-    } catch {*} {
+    } on error {_ eOptions} {
         # This code must run to completion, or the installation might be left
         # in an inconsistent state. We store the old signal handling state,
         # block the critical signals and restore to the previous state instead
@@ -647,11 +727,14 @@ proc _activate_contents {port {rename_list {}}} {
             signal set $osignals
         }
 
+        throw [dict get $eOptions -errorcode] [dict get $eOptions -errorinfo]
+    } finally {
+        #foreach entry [array names todeactivate] {
+        #    registry::entry close $entry
+        #}
         # remove temp image dir
         ::file delete -force $extracted_dir
-        throw
     }
-    ::file delete -force $extracted_dir
 }
 
 # These directories should not be removed during deactivation even if they are empty.
@@ -690,7 +773,7 @@ proc _deactivate_contents {port imagefiles {force 0} {rollback 0}} {
     set files [list]
 
     foreach file $imagefiles {
-        if { [::file exists $file] || (![catch {::file type $file}] && [::file type $file] eq "link") } {
+        if { [::file exists $file] || (![catch {::file type $file} ft] && $ft eq "link") } {
             # Normalize the file path to avoid removing the intermediate
             # symlinks (remove the empty directories instead)
             # Remark: paths in the registry may be not normalized.
@@ -700,7 +783,7 @@ proc _deactivate_contents {port imagefiles {force 0} {rollback 0}} {
             # match and activate will say that some file exists but doesn't
             # belong to any port.
             # The custom realpath proc is necessary because file normalize
-            # does not resolve symlinks on OS X < 10.6
+            # does not resolve symlinks on Mac OS X < 10.6
             set directory [realpath [::file dirname $file]]
             lappend files [::file join $directory [::file tail $file]]
 
@@ -755,5 +838,74 @@ proc _deactivate_contents {port imagefiles {force 0} {rollback 0}} {
     }
 }
 
+# Create a new registry entry using the given metadata array
+proc install {metadata} {
+    global macports::registry.path
+    array set m $metadata
+    registry::write {
+        # store portfile
+        set portfile_sha256 [sha256 file $m(portfile_path)]
+        set portfile_size [file size $m(portfile_path)]
+        set portfile_reg_dir [file join ${registry.path} registry portfiles $m(name)-$m(version)_$m(revision) ${portfile_sha256}-${portfile_size}]
+        set portfile_reg_path ${portfile_reg_dir}/Portfile
+        file mkdir $portfile_reg_dir
+        if {![file isfile $portfile_reg_path] || [file size $portfile_reg_path] != $portfile_size
+                || [sha256 file $portfile_reg_path] ne $portfile_sha256} {
+            file copy -force $m(portfile_path) $portfile_reg_dir
+            file attributes $portfile_reg_path -permissions 0644
+        }
+
+        # store portgroups
+        if {[info exists m(portgroups)]} {
+            foreach {pgname pgversion groupFile} $m(portgroups) {
+                set pgsha256 [sha256 file $groupFile]
+                set pgsize [file size $groupFile]
+                set pg_reg_dir [file join ${registry.path} registry portgroups ${pgsha256}-${pgsize}]
+                set pg_reg_path ${pg_reg_dir}/${pgname}-${pgversion}.tcl
+                lappend portgroups [list $pgname $pgversion $pgsha256 $pgsize]
+                if {![file isfile $pg_reg_path] || [file size $pg_reg_path] != $pgsize || [sha256 file $pg_reg_path] ne $pgsha256} {
+                    file mkdir $pg_reg_dir
+                    file copy -force $groupFile $pg_reg_dir
+                }
+                file attributes $pg_reg_path -permissions 0644
+            }
+            unset m(portgroups)
+        }
+
+        set regref [registry::entry create $m(name) $m(version) $m(revision) $m(variants) $m(epoch)]
+        $regref installtype image
+        $regref state imaged
+        $regref portfile ${portfile_sha256}-${portfile_size}
+        if {[info exists portgroups]} {
+            foreach p $portgroups {
+                $regref addgroup {*}$p
+            }
+        }
+        foreach dep_portname $m(depends) {
+            $regref depends $dep_portname
+        }
+        if {[info exists m(files)]} {
+            # register files
+            $regref map $m(files)
+            unset m(files)
+        }
+        if {[info exists m(binary)]} {
+            foreach {f isbinary} $m(binary) {
+                set fileref [registry::file open [$regref id] $f]
+                $fileref binary $isbinary
+                registry::file close $fileref
+            }
+            unset m(binary)
+        }
+        unset m(name) m(version) m(revision) m(variants) m(epoch) m(depends) \
+              m(portfile_path)
+
+        # remaining metadata maps directly to reg entry fields
+        foreach key [array names m] {
+            $regref $key $m($key)
+        }
+    }
+}
+
 # End of portimage namespace
 }
